package matrixPower;

import java.io.File;
import java.io.IOException;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import matrixPower.gui.Gui;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/* Nacteni matice */
public class MatrixReader {
	
    private Gui gui;
    
        /* XSD schema pro kontrolu struktury matice v XML souboru */
        private final String SCHEMA_FILE = "matrix.xsd";    
    
	public MatrixReader(){
            gui = null;
	}
        
        public MatrixReader(Gui gui){
            this.gui = gui;
	}
	
        /* poskytnuti statickych dat */
	public Matrix load(){
		Matrix matrix = new Matrix(3);
		for (int i = 0; i < 3; i++){
			for (int j = 0; j < 3; j++){
				matrix.set(i, j, (i + j > 0 )? i + j : 1);
			}
		}
		return matrix;
	}
	
        /* Nacteni matice ze zadaneho XML souboru */
        public Matrix loadXML(String sourcePath) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException{
            return loadXML(new File(sourcePath));
        }
        
        /* Nacteni matice ze zadaneho XML souboru */
	public Matrix loadXML(File sourcePath) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException{
		DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
	    domFactory.setNamespaceAware(true); // never forget this!
	    ErrorHandler errorHandler;
            if (gui == null){
                errorHandler = new MyErrorHandler();
            }
            else {
                errorHandler = new MyErrorHandler(gui);
            }
	    
	    SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
            /* kontrola struktury podle schematu */
	    Schema schemaXSD = schemaFactory.newSchema(new File(SCHEMA_FILE));
	    domFactory.setSchema(schemaXSD);
	    DocumentBuilder builder = domFactory.newDocumentBuilder();
	    builder.setErrorHandler(errorHandler);
	    Document doc = builder.parse(sourcePath);

	    XPathFactory factory = XPathFactory.newInstance();
	    XPath xpath = factory.newXPath();
            /* nacteni stupne matice */
	    XPathExpression expr = xpath.compile("/matrix/@grade");
	    Object result = expr.evaluate(doc, XPathConstants.NUMBER);
	    int grade = ((Double) result).intValue();
	    
            /* nacteni radku matice */
	    expr = xpath.compile("//row");
	    result = expr.evaluate(doc, XPathConstants.NODESET);
	    NodeList nodes = (NodeList) result;
	    
	    Matrix matrix = new Matrix(grade);
	    
	    String errMsg = grade + " rows expected. There are " + nodes.getLength() + " rows";
	    if (nodes.getLength() < grade){	    	 
	    	errorHandler.error(new SAXParseException(errMsg, null, null, 0 ,0));
	    }
	    else if (nodes.getLength() > grade){
	    	errorHandler.warning(new SAXParseException(errMsg, null, null, 0 ,0));
	    }
	    
	    int entry;
            /* nacteni prvky radku matice */            
	    for (int i = 0; i < grade; i++) {
	    	expr = xpath.compile("entry/text()");
		    NodeList entries = (NodeList) expr.evaluate(nodes.item(i), XPathConstants.NODESET);
                    /* kontrola, zda pocet radku a sloupcu odpovida stupni matice */
		    if (entries.getLength() < grade){
		    	errMsg = grade + " entries expected. There are " + entries.getLength() + " entries"; 
		    	errorHandler.error(new SAXParseException(errMsg, null, null, 0 ,0));
		    }
		    else if (entries.getLength() > grade){
		    	errMsg = grade + " entries expected. There are " + entries.getLength() + " entries"; 
		    	errorHandler.warning(new SAXParseException(errMsg, null, null, 0 ,0));
		    }
		    for (int j = 0; j < grade; j++){
		    	entry = Integer.parseInt(entries.item(j).getNodeValue());
		    	matrix.set(i, j, entry);
		    }
	    }
	    
	    return matrix;

	}
}
