package matrixPower;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

/* Farmer spoustejici workery a zadavajici jim praci */
public class PowerMonitor {

	/* Trida delnika */
	private class Worker extends Thread {

		private PowerMonitor monitor;    /* reference monitoru */
		private int myId;                /* id delnika */

		public Worker (PowerMonitor par_monitor_ref, int id) {
			monitor = par_monitor_ref;
			myId = id;
		}
		
                /* vypocet jedne souradnice mocniny matice */
		private int multiply(int row, int column){
			int sum = 0;
                        int grade = monitor.getMatrix().getGrade();
                        
                        for (int m = 0; m < grade; m++){
                            sum += monitor.getPowerMatrix().get(row , m) * monitor.getMatrix().get(m, column);
			}
			return sum;
		}

                @Override
                /* prace delnika */
		public void run () {
			WorkPart part;  /* prace, kterou ma delnik spocitat */
			List<Coordinate> locResult = new ArrayList<Coordinate>();   /* spocitane vysledky */
			
			while (monitor.waitForWork(this.myId)) { // blocking
                                Main.addText("Delnik " + myId + " pracuje ..");
                                /* prideleni prace */
				while (!(part = monitor.getWork()).isEmpty()) {
					/* pocitani prvku mocniny matice */
					while (!part.isEmpty()) {
						Coordinate cord = part.getCoordinates().poll();
						cord.setValue(multiply(cord.getRow(), cord.getColumn()));
						locResult.add(cord);
						yield();
					}					
				}	
                                Main.addText("Dokoncil jsem praci: " + this.myId);
                                /* zaslani vysledku rodici */
				monitor.updateResult(locResult, this.myId);
				locResult = new ArrayList<Coordinate>();

			}
                        Main.addText("worker " + myId + " terminated");
		}
	}


	private int workersCount;   /* pocet delniku */
	private int barrierCnt;     /* pocet hotove prace */
	private int cordCnt;        /* pocet spocitanych souradnic */
	private int step;           /* velikost prirazovane prace */
	private int power;          /* stupen mocniny */
	
	private Matrix matrix;      /* puvodni matice */
        private Matrix powerMatrix; /* n-ta mocnina matice */
	private Matrix result;      /* vysledek mocneni matice */
	//private Queue<Coordinate> workQueue;    /* fronta prirazovane prace */
        private Queue<WorkPart> workQueue;    /* fronta prirazovane prace */
	
	private boolean isWork;     /* jestli je nejaka prace */
	private boolean stopWorkers;  /* zda jsou delnici zastaveni */

	private Worker[] workers;    /* pole delniku */


	public PowerMonitor (int n_proc, int step, int power) {
		setWorkersCount(n_proc);
		setStep(step);
		setPower(power);
		isWork = false;
		stopWorkers = false;
		barrierCnt = 0;
		cordCnt = 0;
		
		workers = new Worker[getWorkersCount()];
		for (int i = 0; i < getWorkersCount(); i++) {
			workers[i] = new Worker (this, i);
			workers[i].start();
		}
	}

	/* vypocet mocniny matice */
	public synchronized Matrix power (Matrix matrix) {
		
		this.setMatrix(matrix);
		this.setPowerMatrix(matrix);
		setResult(new Matrix(matrix.getGrade()));
		barrierCnt = 0;
		cordCnt = 0;
		isWork = true;
		
		setWorkParts();
		
		notifyAll();                // it awakes waiting workers

		while (isWork) try {       // wait for final result
			wait ();
		} catch (InterruptedException e) {};

		return getResult();
	}
	
        /* pocatecni generovani fronty prace */
	private void setWorkParts(){
		int grade = this.getMatrix().getGrade();
                workQueue = new LinkedList<WorkPart>();
                WorkPart workPart = new WorkPart();
               
                /* vkladani souradnic do fronty */
		for (int m = 1; m < this.power; m++){
		
                        workPart = new WorkPart();
			for (int i = 0; i < grade; i++){
				for (int j = 0; j < grade; j++){
                                    workPart.addCoordinate(new Coordinate(i, j));
                                    if (workPart.getCoordinates().size() == this.getStep()){
                                        workQueue.add(workPart);
                                        workPart = new WorkPart();
                                    }
				}                                
			}
                        if (!workPart.isEmpty()){
                            workQueue.add(workPart);
                            workPart = new WorkPart();
                        }
                        
                        /* vlozeni prazdne souradnice mezi jednotlivymi mocninami */
			if (m != power - 1){
                            workQueue.add(new WorkPart());
			}					
		}		
		isWork = true;
	}

        /* zastaveni delniku */
	public synchronized int stop () {
		stopWorkers = true;
		notifyAll();
		return 0;
	}

	/* pozastaveni delniku, dokud neni prace */
	private synchronized boolean waitForWork(int id) {
		
		while ((!isWork) && (!stopWorkers)) try {
			wait ();
		} catch (InterruptedException e) {};
		
		if (stopWorkers) {
			return false;  // workers end their working loop
		}
		else {
			return true;  // worker has something to do
		}
	}

        /* prirazeni prace delnikovi, tj. nekolik souradnic */
	private synchronized WorkPart getWork () {
                    
            if ((workQueue.isEmpty()) || (workQueue.peek().isEmpty())){
                return new WorkPart();
            }
            return workQueue.poll();
	}

        /* prijmuti vysledku od delnika */
        /* blokuje delniky, dokud neni dokoncena konkretni mocnina */
	private synchronized void updateResult (List<Coordinate> coordinates, int id) {
				
                Main.addText("Worker " + id + " updates (" + coordinates.size() + " coordinates).");
		cordCnt += coordinates.size();
                Main.addText("\t" + coordinates + ", sum: " + cordCnt);
                /* aktualizovani vysledne matice */
		for (Coordinate cord: coordinates){
			getResult().set(cord);
		}
                
                	
		barrierCnt++;
                Main.addText("barrierCnt: " + barrierCnt + " , thread: " + id);
                
                /* pokud jsou spocitany vsechny souradnice, ukoncit vypocet */
		if (cordCnt == (power - 1) * matrix.getGrade() * matrix.getGrade()){
			isWork = false;
                        notifyAll();
                        return;
		}
		
		/* pokud neposlali vysledky vsichni delnici, uspat delnika */
		while (barrierCnt % getWorkersCount() != 0) {
			try {         
				wait();
			} catch (InterruptedException e) {};
		}
		
		
                /* priprava dalsi mocniny;
                 * nasobit se bude vysledna matice koncici mocniny
                 */
		this.setPowerMatrix(this.getResult());
		
                /* vyjmout prazdnou souradnici z fronty prace, 
                 * aby se mohlo pokracovat v dalsi mocnine */
		if (!this.workQueue.isEmpty() && this.workQueue.peek().isEmpty()){
			this.workQueue.poll();
		}
		
                /* vzbuzeni vsech delniku */
		notifyAll();
		
	}

	public void setResult(Matrix result) {
		this.result = result;
	}

	public Matrix getResult() {
		return result;
	}

	public void setMatrix(Matrix matrix) {
		this.matrix = matrix;
	}

	public Matrix getMatrix() {
		return matrix;
	}

	public void setStep(int step) {
		this.step = step;
	}

	public int getStep() {
		return step;
	}

	public void setWorkersCount(int workersCount) {
		this.workersCount = workersCount;
	}

	public int getWorkersCount() {
		return workersCount;
	}

	public void setPower(int power) {
		this.power = power;
	}

	public int getPower() {
		return power;
	}

        /* nastaveni matice, ktera se bude nasobit */
	private void setPowerMatrix(Matrix powerMatrix) {
            this.powerMatrix = new Matrix(powerMatrix.getGrade());
            for (int i = 0; i < powerMatrix.getGrade(); i++){
                for (int j = 0; j < powerMatrix.getGrade(); j++){
                    this.powerMatrix.set(i, j, powerMatrix.get(i, j));
                }
            }
	}

	private Matrix getPowerMatrix() {
		return powerMatrix;
	}
}
